// $Id:recfriiobs.cpp 5145 2008-05-05 13:44:38Z clworld $

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#include <fcntl.h>

#include <inttypes.h>
#include <dirent.h>
#include <errno.h>

#include <math.h>

#include <err.h>

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <map>

#include <glib.h>

#include <linux/usb/ch9.h>
#include <linux/usbdevice_fs.h>

#include <boost/thread/thread.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/filesystem/operations.hpp>
#include <boost/regex.hpp>
#include <boost/format.hpp>

#include <exception>
#include <stdexcept>

#include "ringbuf.hpp"

// 受信サイズ(usbfsのバルク転送限界サイズ)
static const uint32_t TSDATASIZE = 16384;

// FIFOバッファ設定
static const unsigned long ASYNCBUFFTIME = 10UL; // バッファ長 = 10秒
static const unsigned long ASYNCBUFFSIZE = 0x400000UL * ASYNCBUFFTIME / TSDATASIZE; // 平均32Mbpsとする

// 
static const unsigned int REQUEST_RESERVE_NUM  = 16; // 非同期リクエスト予約数
static const unsigned int REQUEST_POLLING_WAIT = 10; // 非同期リクエストポーリング間隔(ms)

static const unsigned int REQUEST_TIMEOUT    = 1000; // USBリクエストタイムアウト(ms)
static const unsigned int REQUEST_TIMEOUT_TS = 2000; // TS読み込み用USBタイムアウト時間(ms)

static const uint16_t TARGET_ID_VENDOR  = 0x7a69; // FriioのidVendor
static const uint16_t TARGET_ID_PRODUCT = 0x0001; // FriioのidProduct

static const uint32_t TIMEOUT_RETRY_MAX      = 3;    // 初期化時にタイムアウトエラーが起こった場合のリトライ回数
static const uint32_t TIMEOUT_RETRY_INTERVAL = 1000; // ↑の間隔(ms)

static const uint32_t SETCHANNEL_COMMAND_INTERBAL = 5; // SetChannel時Timeout error回避用遅延時間(ms)
static const int TARGET_INTERFACE = 0;
static const int TARGET_ENDPOINT  = 0x81;

// LED色
static const int BON_LED_PURPLE = 0; // 紫(待機中)
static const int BON_LED_GREEN  = 1; // 緑(視聴中)
static const int BON_LED_RED    = 2; // 赤(録画中)
static const int BON_LED_WHITE  = 3; // 白(エラー、送信データサンプルがない＞、＜)

typedef enum {
	FRIIO_WHITE,
	FRIIO_BLACK,
	FRIIO_TYPES_LEN,
} FriioType;

static std::string friioTypeName[FRIIO_TYPES_LEN] = { "Friio(White)", "Friio(Black)" };

//
static const float SIGNALLEVEL_RETRY_THRESHOLD = 10.0; // 開始時の信号レベル取得で信号レベルがこれ以下であれば再度信号レベルを取得する。
static const int   SIGNALLEVEL_RETRY_MAX       = 2;    // 信号レベル再取得の最大回数
static const int   SIGNALLEVEL_RETRY_INTERVAL  = 400;  // 信号レベル再取得間隔(ms)

static const char *BASE_DIR_UDEV  = "/dev/bus/usb";  // udevで作成されるUSBのディレクトリ
static const char *BASE_DIR_USBFS = "/proc/bus/usb"; // usbfsのディレクトリ

static const char *DEVICES_UDEV  = "/dev/bus/usb/devices";  // udevのdevicesファイルパス(恐らく存在しないor間違っているが念の為)
static const char *DEVICES_USBFS = "/proc/bus/usb/devices"; // usbfsのdevicesファイルパス

static int tuner_ctrl(int fd, usbdevfs_ctrltransfer *ctrl);
static void tuner_claim(int fd, unsigned int interface);
static void tuner_release(int fd, unsigned int interface);
static bool is_friio(const std::string &devfile);
static int  open_tuner();

static void UsbProcStreamInit(int fd);
static void UsbProcLED(int fd, uint8_t bType);
static void UsbProcFixInitB(int fd);
static FriioType UsbProcFixInitA(int fd);
static void UsbProcBegin(int fd);
static void UsbProcEnd(int fd);
static void UsbTerminate(int fd);
static void UsbInitalize(int fd);

static void SetChannel(int fd, const uint8_t bCh);
static void SetChannel_Black(int fd, const uint32_t dwSpace, const uint32_t dwChannel);
static float GetSignalLevel(int fd);
static int tuner_readts(int fd, uint8_t *buf, int len);

// USBコマンドのタイムアウト時に投げられます。
class timeout_error : public std::runtime_error {
public:
	timeout_error(const std::string&  __arg) : std::runtime_error(__arg) {
	};
};

struct AsyncRequest {
	usbdevfs_urb urb;
	int fd;
	uint8_t buf[TSDATASIZE];
};

static void discard_async_request(AsyncRequest *r) {
	std::cout << "discards timeouted URB." << std::endl;
	int ret = ioctl(r->fd, USBDEVFS_DISCARDURB, &(r->urb));
	if (ret < 0) {
		err(1, "can't discard urb (at timeout)");
	}
}

// 非同期リクエスト発行スレッド
class PushIoThread {
public:
	PushIoThread(int fd, RingBuf<AsyncRequest> *buf) : fd(fd), buf(buf), is_cancelled(false) {
	}
	virtual ~PushIoThread() {
	}
	
	// スレッド処理本体
	void operator()() {
		while(!is_cancelled) {
			// リクエスト待ちが一杯であればREQUEST_POLLING_WAIT(ms)待ってやりなおし。
			if (REQUEST_RESERVE_NUM <= buf->getWaitCount()) {
				// リクエスト待ちが一杯です。
				usleep(REQUEST_POLLING_WAIT * 1000);
				continue;
			}
			
			// リクエスト待ちに空きがある。
			AsyncRequest *req          = buf->getPushPtr(REQUEST_TIMEOUT_TS, discard_async_request);
			
			memset(&(req->urb), '\0', sizeof(usbdevfs_urb));
			memset(req->buf,    '\0', TSDATASIZE);
			
			req->urb.type              = USBDEVFS_URB_TYPE_BULK;
			req->urb.endpoint          = TARGET_ENDPOINT;
			req->urb.status            = 0;
			req->urb.buffer            = req->buf;
			req->urb.buffer_length     = TSDATASIZE;
			req->urb.actual_length     = 0;
			req->urb.number_of_packets = 0;
			req->urb.signr             = 0;
			req->urb.usercontext       = req;
			req->fd = fd;
			
			int r = ioctl(fd, USBDEVFS_SUBMITURB, &(req->urb));
			if (r < 0) {
				err(1, "can't send urb");
			}
		}
	}
	
	// 実行停止用
	void cancel() {
		is_cancelled = true;
	}
	
protected:
	// URB送信対象fd
	int fd;
	// リングバッファ
	RingBuf<AsyncRequest> *buf;
	// 停止フラグon?
	bool is_cancelled;
};

// 非同期リクエスト受信スレッド
class PopIoThread {
public:
	PopIoThread(int fd, RingBuf<AsyncRequest> *buf) : fd(fd), buf(buf), is_cancelled(false) {
		int pipefds[2];
		int r = pipe(pipefds);
		if (r < 0) {
			err(1, "can't create pipe.");
		}
		
		event_fd_pop  = pipefds[0];
		event_fd_push = pipefds[1];
	}
	virtual ~PopIoThread() {
		if (close(event_fd_push) < 0) {
			err(1, "can't clsoe event_fd_push.");
		}
		if (close(event_fd_pop) < 0) {
			err(1, "can't clsoe event_fd_pop.");
		}
	}
	
	// スレッド処理本体
	void operator()() {
		fd_set readfds, writefds;
		
		while(!is_cancelled) {
			FD_ZERO(&readfds);
			FD_SET(event_fd_pop, &readfds);

			FD_ZERO(&writefds);
			FD_SET(fd, &writefds);
			
			int maxfd = MAX(fd, event_fd_pop);
			int r = select(maxfd+1, &readfds, &writefds, NULL, NULL);
			if (r < 0) {
				err(1, "select failed.");
			}
			
			if (FD_ISSET(event_fd_pop, &readfds)) {
				// cancelから起こされた？
				uint8_t dummybuf[1];
				int r = read(event_fd_pop, dummybuf, 1);
				if (r < 0) {
					err(1, "can't read from event_fd.");
				}
			}
			
			if (FD_ISSET(fd, &writefds)) {
				// usbfsから何か来た
				reapUrbs();
			}
		}
	}
	
	// 実行停止用
	void cancel() {
		is_cancelled = true;
		// selectを起こす為に書き込む
		int r = write(event_fd_push, "n", 1);
		if (r < 0) {
			err(1, "can't write to event_fd.");
		}
	}
	
protected:
	// usbfsから完了URBを読み込む
	void reapUrbs() {
		usbdevfs_urb *urb;
		while (ioctl(fd, USBDEVFS_REAPURBNDELAY, (void *)&urb) >= 0) {
			// 読み込めた
			AsyncRequest *req = reinterpret_cast<AsyncRequest *>(urb->usercontext);
			buf->setReady(req);
		}
	}
	
	// URB受信対象fd
	int fd;
	// リングバッファ
	RingBuf<AsyncRequest> *buf;
	// 停止フラグon?
	bool is_cancelled;
	// select中断用pipe
	int event_fd_push;
	int event_fd_pop;
};

int main(int argc, char *argv[]) {
	if (argc != 4) {
		fprintf(stderr, "usage: %s channel recsec destfile\n", argv[0]);
		fprintf(stderr, "Channels:\n");
		fprintf(stderr, "  13 : BS-Asahi      24 : 110CS #1\n");
		fprintf(stderr, "  14 : BS-i          25 : 110CS #2\n");
		fprintf(stderr, "  15 : WOWOW         26 : 110CS #3\n");
		fprintf(stderr, "  16 : BS-Japan      27 : 110CS #4\n");
		fprintf(stderr, "  17 : BS11          28 : 110CS #5\n");
		fprintf(stderr, "  18 : Star Channel  29 : 110CS #6\n");
		fprintf(stderr, "  19 : TwellV        30 : 110CS #7\n");
		fprintf(stderr, "  20 : BS-NTV        31 : 110CS #8\n");
		fprintf(stderr, "  21 : BS-Fuji       32 : 110CS #9\n");
		fprintf(stderr, "  22 : NHK BS1/BS2   33 : 110CS #a\n");
		fprintf(stderr, "  23 : NHK hi        34 : 110CS #b\n");
		fprintf(stderr, "                     35 : 110CS #c\n");
		exit(1);
	}
	
	int chnum  = atoi(argv[1]);
	int recsec = atoi(argv[2]);
	char *destfile = argv[3];
	
	if (chnum < 13 || 35 < chnum) {
		fprintf(stderr, "channel must be (13 <= channel <= 35).\n");
		exit(1);
	}
	
	if (recsec <= 0) {
		fprintf(stderr, "recsec must be (recsec > 0).\n");
		exit(1);
	}
	
	// 録画時間の基準開始時間
	time_t time_start = time(NULL);

	int fd = open_tuner();

	FILE *dest = fopen(destfile, "w");
	if (NULL == dest) {
		fprintf(stderr, "can't open file '%s' to write.\n", destfile);
		exit(1);
	}
	
	// 
	bool retry = true;
	
	// ファイル出力までの間にTimeoutエラーが発生した場合、USBの初期化からやりなおす。
	retry = true;
	uint32_t retry_count = 0;
	while (retry) {
		retry = false;
		
		try {
			UsbInitalize(fd);
			
			SetChannel(fd, chnum);
			
			float lev_before = GetSignalLevel(fd);
			printf("Signal level: %f\n", lev_before);
			int lev_retry_count = 0;
			while (lev_before < SIGNALLEVEL_RETRY_THRESHOLD && lev_retry_count < SIGNALLEVEL_RETRY_MAX) {
				usleep(SIGNALLEVEL_RETRY_INTERVAL * 1000);
				lev_retry_count++;
				lev_before = GetSignalLevel(fd);
				printf("Signal level: %f\n", lev_before);
			}
		} catch (timeout_error &err) {
			// Timeout -> retry.
			if (TIMEOUT_RETRY_MAX <= retry_count) {
				errx(1, "Timeout error. abort.");
			}
			
			std::cout << "Timeout error. retry." << std::endl;
			retry_count++;
			close(fd); // Tunerはリトライ時開き直す。
			usleep(TIMEOUT_RETRY_INTERVAL * 1000);
			fd = open_tuner(); // Tunerはリトライ時開き直す。
			retry = true;
		}
	}
	
	printf("Output ts file.\n");
	
	timeval tv_start;
	if (gettimeofday(&tv_start, NULL) < 0) {
		err(1, "gettimeofday failed.");
	}
	
	RingBuf<AsyncRequest> buf(ASYNCBUFFSIZE);
	PushIoThread push_func(fd, &buf);
	PopIoThread  pop_func(fd,  &buf);
	boost::thread push_thread(boost::ref(push_func));
	boost::thread pop_thread(boost::ref(pop_func));
	
	while(time(NULL) <= time_start + recsec) {
		const AsyncRequest *req = buf.getPopPtr(200);
		if (NULL == req) {
			continue;
		}
		
		if (req->urb.status != 0) {
			std::cout << "UrbStatus: " << req->urb.status << std::endl;
			continue;
		}
		
		int rlen = req->urb.actual_length;
		if (0 == rlen) {
			continue;
		}
		
		fwrite(req->buf, 1, rlen, dest);
	}

	timeval tv_end;
	if (gettimeofday(&tv_end, NULL) < 0) {
		err(1, "gettimeofday failed.");
	}
	
	push_func.cancel();
	pop_func.cancel();
	push_thread.join();
	pop_thread.join();
	
	//
	fflush(dest);
	fclose(dest);
	printf("done.\n");
	
	timeval rec_time;
	timersub(&tv_end, &tv_start, &rec_time);
	printf("Rec time: %d.%06d sec.\n", (int)rec_time.tv_sec, (int)rec_time.tv_usec);
	
	try {
		float lev_after = GetSignalLevel(fd);
		printf("Signal level: %f\n", lev_after);
		
		UsbTerminate(fd);
		
		tuner_release(fd, TARGET_INTERFACE);
	} catch (timeout_error &err) {
		std::cout << "Timeout error. ignored." << std::endl;
		// USBのタイムアウトエラーがあった場合無視して終了する。
	}
	
	int r = close(fd);
	if (-1 == r) {
		err(1, "can't close fd.");
	}
	//////////////////////////////////////////////////////////////////////////
}

// Request=0x40 RequestNo=0x01 で Length=0 のベンダリクエストを送信
void UsbSendVendorRequest(int fd, uint16_t wValue, uint16_t wIndex)
{
	uint8_t buf[] = "";
	
	usbdevfs_ctrltransfer ctrl;
	ctrl.bRequestType = USB_DIR_OUT|USB_TYPE_VENDOR|USB_RECIP_DEVICE;
	ctrl.bRequest     = 0x01;
	ctrl.wValue       = wValue;
	ctrl.wIndex       = wIndex;
	ctrl.wLength      = 0;
	ctrl.timeout      = REQUEST_TIMEOUT;
	ctrl.data         = buf;
	
	tuner_ctrl(fd, &ctrl);
}

int tuner_readts(int fd, uint8_t *buf, int len) {
	usbdevfs_bulktransfer bulk;
	bulk.ep      = TARGET_ENDPOINT;
	bulk.len     = len;
	bulk.timeout = REQUEST_TIMEOUT;
	bulk.data    = buf;
	
	int r = ioctl(fd, USBDEVFS_BULK, &bulk);
	if (-1 == r) {
		err(1, "usb bulk failed.");
	}
	return r;
}

int tuner_ctrl(int fd, usbdevfs_ctrltransfer *ctrl) {
	int r = ioctl(fd, USBDEVFS_CONTROL, ctrl);
	if (-1 == r) {
		if (ETIMEDOUT == errno) {
			throw timeout_error("tuner_ctrl: timeout.");
		} else {
			err(1, "usb ctrl failed.");
		}
	}
	return r;
}

void tuner_claim(int fd, unsigned int interface) {
	int r = ioctl(fd, USBDEVFS_CLAIMINTERFACE, &interface);
	if (-1 == r) {
		err(1, "can't claim if 0.");
	}
}

void tuner_release(int fd, unsigned int interface) {
	int r = ioctl(fd, USBDEVFS_RELEASEINTERFACE, &interface);
	if (-1 == r) {
		err(1, "can't release if 0.");
	}
}

void SetChannel(int fd, const uint8_t bCh)
{
	if ((bCh >= 13U) && (bCh <= 23U)) {
		// BS
		SetChannel_Black(fd, 0U, bCh - 13U);
	} else if ((bCh >= 24U) && (bCh <= 35U)){
		// 110CS
		SetChannel_Black(fd, 1U, bCh - 24U);
	} else {
		// 非対応
		errx(1, "unknown channel %d", bCh);
	}
}

void SetChannel_Black(int fd, const uint32_t dwSpace, const uint32_t dwChannel)
{
	// 資料不足のため未実装
	struct Black_Channel_Sequence {
		uint8_t byReqNo;
		uint16_t wValue;
		uint16_t wIndex;
		uint8_t bData;		// 3byteのデータ部があるか
		uint8_t byData1;	// データ１
		uint8_t byData2;	// データ２
		uint8_t byData3;	// データ３
	};
	
	struct Black_Channel_Sequence cmd[] = {
														// Index(n回目-1)
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x00, 0x0A },	// 0
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x01, 0x4C },	// 1
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x02, 0x02 },	// 2
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x03, 0x40 },	// 3
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x04, 0x6C },	// 4
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x05, 0x47 },	// 5
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x06, 0x02 },	// 6
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x07, 0x30 },	// 7
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x08, 0x20 },	// 8
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x09, 0xA2 },	// 9
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0A, 0xF1 },	// 10
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0B, 0x38 },	// 11
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0C, 0xD0 },	// 12
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0D, 0x10 },	// 13
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0E, 0xB0 },	// 14
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0F, 0x01 },	// 15
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x00, 0x0A },	// 16
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x01, 0x4C },	// 17
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x02, 0x02 },	// 18
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x03, 0x40 },	// 19
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x04, 0x6C },	// 20
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x05, 0x47 },	// 21
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x06, 0x02 },	// 22
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x07, 0x30 },	// 23
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x08, 0x20 },	// 24
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x09, 0xA2 },	// 25
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0A, 0xF1 },	// 26
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0B, 0x38 },	// 27
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0C, 0xD0 },	// 28
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0D, 0x10 },	// 29
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0E, 0xB0 },	// 30
		{ 0x03, 0x3200, 0x00FE, TRUE, 0xC0, 0x0F, 0x01 },	// 31
		{ 0x01, 0x3210, 0x0001, FALSE, 0x00, 0x00, 0x00 },	// 32
		{ 0x01, 0x3200, 0x0002, FALSE, 0x00, 0x00, 0x00 },	// 33
		{ 0x01, 0x3200, 0x0003, FALSE, 0x00, 0x00, 0x00 },	// 34
		{ 0x01, 0x3200, 0x0008, FALSE, 0x00, 0x00, 0x00 },	// 35
		{ 0x01, 0x3259, 0x000C, FALSE, 0x00, 0x00, 0x00 },	// 36
		{ 0x01, 0x32F2, 0x000D, FALSE, 0x00, 0x00, 0x00 },	// 37
		{ 0x01, 0x3251, 0x000F, FALSE, 0x00, 0x00, 0x00 },	// 38
		{ 0x01, 0x32B2, 0x0010, FALSE, 0x00, 0x00, 0x00 },	// 39
		{ 0x01, 0x3204, 0x001A, FALSE, 0x00, 0x00, 0x00 },	// 40
		{ 0x01, 0x3204, 0x001B, FALSE, 0x00, 0x00, 0x00 },	// 41
		{ 0x01, 0x3240, 0x0038, FALSE, 0x00, 0x00, 0x00 },	// 42
		{ 0x01, 0x3210, 0x0039, FALSE, 0x00, 0x00, 0x00 },	// 43
		{ 0x01, 0x3208, 0x003C, FALSE, 0x00, 0x00, 0x00 },	// 44
		{ 0x01, 0x3296, 0x003D, FALSE, 0x00, 0x00, 0x00 },	// 45
		{ 0x01, 0x32D9, 0x003E, FALSE, 0x00, 0x00, 0x00 },	// 46
		{ 0x01, 0x32B6, 0x003F, FALSE, 0x00, 0x00, 0x00 },	// 47
		{ 0x01, 0x325D, 0x0040, FALSE, 0x00, 0x00, 0x00 },	// 48
		{ 0x01, 0x3299, 0x0041, FALSE, 0x00, 0x00, 0x00 },	// 49
		{ 0x01, 0x3258, 0x0044, FALSE, 0x00, 0x00, 0x00 },	// 50
		{ 0x01, 0x3254, 0x0045, FALSE, 0x00, 0x00, 0x00 },	// 51
		{ 0x01, 0x3230, 0x0046, FALSE, 0x00, 0x00, 0x00 },	// 52
		{ 0x01, 0x3241, 0x0047, FALSE, 0x00, 0x00, 0x00 },	// 53
		{ 0x01, 0x3216, 0x0049, FALSE, 0x00, 0x00, 0x00 },	// 54
		{ 0x01, 0x3241, 0x004A, FALSE, 0x00, 0x00, 0x00 },	// 55
		{ 0x01, 0x3241, 0x004B, FALSE, 0x00, 0x00, 0x00 },	// 56
		{ 0x01, 0x325E, 0x004F, FALSE, 0x00, 0x00, 0x00 },	// 57
		{ 0x01, 0x3204, 0x0050, FALSE, 0x00, 0x00, 0x00 },	// 58
		{ 0x01, 0x32C0, 0x0051, FALSE, 0x00, 0x00, 0x00 },	// 59
		{ 0x01, 0x328A, 0x0052, FALSE, 0x00, 0x00, 0x00 },	// 60
		{ 0x01, 0x3213, 0x0053, FALSE, 0x00, 0x00, 0x00 },	// 61
		{ 0x01, 0x3290, 0x0054, FALSE, 0x00, 0x00, 0x00 },	// 62
		{ 0x01, 0x3292, 0x0055, FALSE, 0x00, 0x00, 0x00 },	// 63
		{ 0x01, 0x326D, 0x0058, FALSE, 0x00, 0x00, 0x00 },	// 64
		{ 0x01, 0x326F, 0x0059, FALSE, 0x00, 0x00, 0x00 },	// 65
		{ 0x01, 0x322E, 0x005A, FALSE, 0x00, 0x00, 0x00 },	// 66
		{ 0x01, 0x3223, 0x005B, FALSE, 0x00, 0x00, 0x00 },	// 67
		{ 0x01, 0x3233, 0x005D, FALSE, 0x00, 0x00, 0x00 },	// 68
		{ 0x01, 0x3204, 0x0087, FALSE, 0x00, 0x00, 0x00 },	// 69
		{ 0x01, 0x3200, 0x008D, FALSE, 0x00, 0x00, 0x00 },	// 70
		{ 0x01, 0x3000, 0x0077, FALSE, 0x00, 0x00, 0x00 },	// 71
		{ 0x01, 0x3000, 0x0071, FALSE, 0x00, 0x00, 0x00 },	// 72
		{ 0x01, 0x3000, 0x0071, FALSE, 0x00, 0x00, 0x00 },	// 73
		{ 0x01, 0x3240, 0x008F, FALSE, 0x00, 0x00, 0x00 },	// 74
		{ 0x01, 0x32F2, 0x0090, FALSE, 0x00, 0x00, 0x00 },	// 75
		{ 0x01, 0x3207, 0x00A5, FALSE, 0x00, 0x00, 0x00 },	// 76
		{ 0x01, 0x3207, 0x00A5, FALSE, 0x00, 0x00, 0x00 },	// 77
		{ 0x01, 0x32FF, 0x00A3, FALSE, 0x00, 0x00, 0x00 },	// 78
		{ 0x01, 0x32F7, 0x00A3, FALSE, 0x00, 0x00, 0x00 },	// 79
		{ 0x01, 0x3040, 0x0001, FALSE, 0x00, 0x00, 0x00 },	// 80
		{ 0x01, 0x3030, 0x0004, FALSE, 0x00, 0x00, 0x00 },	// 81
		{ 0x01, 0x3040, 0x0005, FALSE, 0x00, 0x00, 0x00 },	// 82
		{ 0x01, 0x3040, 0x0007, FALSE, 0x00, 0x00, 0x00 },	// 83
		{ 0x01, 0x30FF, 0x000F, FALSE, 0x00, 0x00, 0x00 },	// 84
		{ 0x01, 0x3027, 0x0011, FALSE, 0x00, 0x00, 0x00 },	// 85
		{ 0x01, 0x302F, 0x0013, FALSE, 0x00, 0x00, 0x00 },	// 86
		{ 0x01, 0x3031, 0x0014, FALSE, 0x00, 0x00, 0x00 },	// 87
		{ 0x01, 0x3002, 0x0016, FALSE, 0x00, 0x00, 0x00 },	// 88
		{ 0x01, 0x3020, 0x0022, FALSE, 0x00, 0x00, 0x00 },	// 89
		{ 0x01, 0x3070, 0x0024, FALSE, 0x00, 0x00, 0x00 },	// 90
		{ 0x01, 0x3079, 0x002C, FALSE, 0x00, 0x00, 0x00 },	// 91
		{ 0x01, 0x3034, 0x002D, FALSE, 0x00, 0x00, 0x00 },	// 92
		{ 0x01, 0x3028, 0x0030, FALSE, 0x00, 0x00, 0x00 },	// 93
		{ 0x01, 0x3031, 0x0031, FALSE, 0x00, 0x00, 0x00 },	// 94
		{ 0x01, 0x3001, 0x0038, FALSE, 0x00, 0x00, 0x00 },	// 95
		{ 0x01, 0x3078, 0x0039, FALSE, 0x00, 0x00, 0x00 },	// 96
		{ 0x01, 0x3090, 0x0048, FALSE, 0x00, 0x00, 0x00 },	// 97
		{ 0x01, 0x3068, 0x0051, FALSE, 0x00, 0x00, 0x00 },	// 98
		{ 0x01, 0x3038, 0x005E, FALSE, 0x00, 0x00, 0x00 },	// 99
		{ 0x01, 0x3021, 0x00C0, FALSE, 0x00, 0x00, 0x00 },	// 100
		{ 0x01, 0x3010, 0x00C1, FALSE, 0x00, 0x00, 0x00 },	// 101
		{ 0x01, 0x301A, 0x00E4, FALSE, 0x00, 0x00, 0x00 },	// 102
		{ 0x01, 0x301F, 0x00EA, FALSE, 0x00, 0x00, 0x00 },	// 103
		{ 0x00, 0x0000, 0x0000, FALSE, 0x00, 0x00, 0x00 },	// End
	};
	
	// チューニングシーケンス可変部分
	static const uint8_t PllConfA[] = {0x08, 0x08, 0x08, 0x08, 0x09, 0x09, 0x09, 0x0A, 0x0A, 0x0A, 0x0A, 0x0C, 0x0C, 0x0D, 0x0D, 0x0D, 0x0E, 0x0E, 0x0E, 0x0F, 0x0F, 0x0F, 0x10};	// 1/17回目 +40
	static const uint8_t PllConfB[] = {0x32, 0x32, 0x80, 0x80, 0x64, 0x64, 0x64, 0x00, 0x00, 0x4C, 0x4C, 0x9A, 0xEA, 0x3A, 0x8A, 0xDA, 0x2A, 0x7A, 0xCA, 0x1A, 0x6A, 0xBA, 0x0A};	// 2/18回目 +40
	static const uint8_t PllConfC[] = {0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};	// 75回目	+02
	static const uint8_t PllConfD[] = {0x10, 0x11, 0x30, 0x31, 0x90, 0x91, 0x92, 0xD0, 0xD1, 0xF1, 0xF2, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};	// 76回目	+02
	
	const uint32_t dwChIndex = dwSpace * 11UL + dwChannel;
	
	// 資料
	// BSAT-2c/3a TsID=4[BSch番号(2byte)][TP番号]
	// BS- 1  周波数=11.727GHz 0832 TsID=4010 BS朝日
	// BS- 1  周波数=11.727GHz 0832 TsID=4011 BS-i
	// BS- 3  周波数=11.765GHz 0880 TsID=4030 WOWWOW
	// BS- 3  周波数=11.765GHz 0880 TsID=4031 BSジャパン
	// BS- 5                                   アナログで使用(WOWWOW)
	// BS- 7                                   アナログで使用(NHK BS-1)
	// BS- 9  周波数=11.880GHz 0964 TsID=4090 BS11
	// BS- 9  周波数=11.880GHz 0964 TsID=4091 スターチャンネル
	// BS- 9  周波数=11.880GHz 0964 TsID=4092 TwellV
	// BS-11                                   アナログで使用(NHK BS-2)
	// BS-13  周波数=11.957GHz 0A00 TsID=40D0 BS日テレ
	// BS-13  周波数=11.957GHz 0A00 TsID=40D1 BSフジ
	// BS-15  周波数=11.996GHz 0A4C TsID=40F1 NHK BS1/BS2/ウェザーニューズ/エンジニアリングストリーム
	// BS-15  周波数=11.996GHz 0A4C TsID=40F2 NHK BS-Hi
	// BS-17                                   未使用(BSAT-4以降に搭載予定)
	// BS-19                                   未使用(BSAT-4以降に搭載予定)
	// BS-21                                   未使用(BSAT-4以降に搭載予定)
	// BS-23                                   未使用(BSAT-4以降に搭載予定)
	//
	// NSAT-110 TsID=0000 固定
	// CSND2  周波数=12.291GHz 0C9A TsID=0000 CS
	// CSND4  周波数=12.331GHz 0CEA TsID=0000 CS
	// CSND6  周波数=12.371GHz 0D3A TsID=0000 CS
	// CSND8  周波数=12.411GHz 0D8A TsID=0000 CS
	// CSND10 周波数=12.451GHz 0DDA TsID=0000 CS
	// CSND12 周波数=12.491GHz 0E2A TsID=0000 CS
	// CSND14 周波数=12.531GHz 0E7A TsID=0000 CS
	// CSND16 周波数=12.571GHz 0ECA TsID=0000 CS
	// CSND18 周波数=12.611GHz 0F1A TsID=0000 CS
	// CSND20 周波数=12.651GHz 0F6A TsID=0000 CS
	// CSND22 周波数=12.691GHz 0FBA TsID=0000 CS
	// CSND24 周波数=12.731GHz 100A TsID=0000 CS
	
	uint32_t dwXmitNum = 0;
	for (;;dwXmitNum++) {
		struct Black_Channel_Sequence tmp = cmd[dwXmitNum];
		if (tmp.byReqNo == 0x00) break; // 終端
		
		// リクエストの作成
		uint8_t  buf[3]  = "";
		uint32_t buf_len = 0;
		
		if (tmp.bData) {
			buf_len = 3;
			buf[0]  = tmp.byData1;
			buf[1]  = tmp.byData2;
			buf[2]  = tmp.byData3;
		}
		
		usbdevfs_ctrltransfer ctrl;
		ctrl.bRequestType = USB_DIR_OUT|USB_TYPE_VENDOR|USB_RECIP_DEVICE;
		ctrl.bRequest     = tmp.byReqNo;
		ctrl.wValue       = tmp.wValue;
		ctrl.wIndex       = tmp.wIndex;
		ctrl.wLength      = buf_len;
		ctrl.timeout      = REQUEST_TIMEOUT;
		ctrl.data         = buf;
		
		// 可変部の変更
		switch(dwXmitNum){
			case 0:		buf[2] = PllConfA[dwChIndex];	break;	// 周波数1 データの最終byteにいれる
			case 1:		buf[2] = PllConfB[dwChIndex];	break;	// 周波数2 データの最終byteにいれる
			case 16:	buf[2] = PllConfA[dwChIndex];	break;	// 周波数1 データの最終byteにいれる
			case 17:	buf[2] = PllConfB[dwChIndex];	break;	// 周波数2 データの最終byteにいれる
			case 74:	ctrl.wValue = (ctrl.wValue & 0xFF00) | ((uint16_t)PllConfC[dwChIndex]);		break;	// TsID1 wValueの下位Byteに入れる
			case 75:	ctrl.wValue = (ctrl.wValue & 0xFF00) | ((uint16_t)PllConfD[dwChIndex]);		break;	// TsID2 wValueの下位Byteに入れる
		}
		
		// 送信
		try {
			tuner_ctrl(fd, &ctrl);
		} catch (timeout_error &err) {
			std::cout << "SetChannel_Black: timeout error at command " << dwXmitNum << std::endl;
			throw;
		}
		
		// 何故かここで遅延をかけないとTimeout errorが起こる。
		usleep(SETCHANNEL_COMMAND_INTERBAL * 1000);
	}
}

float GetSignalLevel(int fd) {
	// 線形補完で近似する
	static const float afLevelTable[] =
	{
		24.07f,	// 00	00	0		24.07dB
		24.07f,	// 10	00	4096	24.07dB
		18.61f,	// 20	00	8192	18.61dB
		15.21f,	// 30	00	12288	15.21dB
		12.50f,	// 40	00	16384	12.50dB
		10.19f,	// 50	00	20480	10.19dB
		8.140f,	// 60	00	24576	8.140dB
		6.270f,	// 70	00	28672	6.270dB
		4.550f,	// 80	00	32768	4.550dB
		3.730f,	// 88	00	34816	3.730dB
		3.630f,	// 88	FF	35071	3.630dB
		2.940f,	// 90	00	36864	2.940dB
		1.420f,	// A0	00	40960	1.420dB
		0.000f	// B0	00	45056	-0.01dB
	};
	
	uint8_t sigbuf[4];
	memset(sigbuf, '\0', sizeof(sigbuf));
	
	// 1
	usbdevfs_ctrltransfer ctrl;
	ctrl.bRequestType = USB_DIR_IN|USB_TYPE_VENDOR|USB_RECIP_DEVICE;
	ctrl.bRequest     = 0x02;
	ctrl.wValue       = 0x3200;
	ctrl.wIndex       = 0x00BA;
	ctrl.wLength      = 4;
	ctrl.timeout      = REQUEST_TIMEOUT;
	ctrl.data         = sigbuf;
	
	// 信号レベル要求
	tuner_ctrl(fd, &ctrl);

	// 信号レベル計算
	if(sigbuf[2] <= 0x10U){
		// 最大クリップ
		return 24.07f;
	} else if (sigbuf[2] >= 0xB0U) {
		// 最小クリップ
		return 0.0f;
	} else {
		// 線形補完
		const float fMixRate = (float)(((uint16_t)(sigbuf[2] & 0x0FU) << 8) | (uint16_t)sigbuf[3]) / 4095.0f;
		return afLevelTable[sigbuf[2] >> 4] * (1.0f - fMixRate) + afLevelTable[(sigbuf[2] >> 4) + 0x01U] * fMixRate;
	}
}

// 白凡黒凡判別
void scan_tuner_type(std::map<std::string, FriioType> *result) {
	boost::filesystem::path devices_file = boost::filesystem::path(DEVICES_UDEV);
	if (!boost::filesystem::exists(devices_file)) {
		devices_file = boost::filesystem::path(DEVICES_USBFS);
	}
	
	std::cout << "devices file: " << devices_file.string() << std::endl;
	
	std::ifstream in;
	in.exceptions(std::ifstream::badbit);
	in.open(devices_file.string().c_str(), std::ios::in);
	
	// hub
	std::vector<std::string> hublist;
	// hub -> Friio種別
	std::map<std::string, FriioType> hub2type;
	// hub -> devices*
	std::map<std::string, std::vector<std::string> > hub2devs;
	
	std::string bus    = "";
	std::string parent = "";
	std::string devno  = "";
	
	boost::regex regex_t("Bus=([0-9]+) .*Prnt=([0-9]+) .*Dev#= *([0-9]+)");
	boost::regex regex_p("Vendor=([A-Fa-f0-9]+) .*ProdID=([A-Fa-f0-9]+)");
	
	std::string line;
	boost::smatch match;
	while (std::getline(in, line)) {
		
		if (line.empty()) {
			continue;
		} else if (line.at(0) == 'T') {
			// トポロジ
			if (boost::regex_search(line, match, regex_t)) {
				bus    = (boost::format("%03d") % std::atoi(match.str(1).c_str())).str();
				parent = (boost::format("%03d") % std::atoi(match.str(2).c_str())).str();
				devno  = (boost::format("%03d") % std::atoi(match.str(3).c_str())).str();
			}
		} else if (line.at(0) == 'P') {
			if (boost::regex_search(line, match, regex_p)) {
				std::string vendor  = match.str(1);
				std::string product = match.str(2);
				
				std::string hub = bus + "/" + parent;
				std::string dev = bus + "/" + devno;
				
				if (vendor == "058f" && product == "9522") {
					hublist.push_back(hub);
					hub2type[hub] = FRIIO_WHITE;
				} else if (vendor == "058f" && product == "9520") {
					hublist.push_back(hub);
					hub2type[hub] = FRIIO_BLACK;
				}
				
				hub2devs[hub].push_back(dev);
			}
		}
	}
	in.close();
	
	for (std::vector<std::string>::iterator iter = hublist.begin(); iter != hublist.end(); ++iter) {
		std::string hub = *iter;
		FriioType friio_type = hub2type[hub];
		
		for (std::vector<std::string>::iterator deviter = hub2devs[hub].begin(); deviter != hub2devs[hub].end(); ++deviter) {
			(*result)[*deviter] = friio_type;
		}
	}
}

int open_tuner() {
	boost::filesystem::path base_dir = boost::filesystem::path(BASE_DIR_UDEV);
	if (!boost::filesystem::is_directory(base_dir)) {
		base_dir = boost::filesystem::path(BASE_DIR_USBFS);
	}
	
	std::cout << "Search friios from dir: " << base_dir << std::endl;
	
	FriioType target_type = FRIIO_BLACK;
	
	// 白黒判別用
	std::map<std::string, FriioType> tuner_type_map;
	scan_tuner_type(&tuner_type_map);
	
	std::vector<std::string> friios;
	// base_dirからfriioをデバイスファイルを探す。
	boost::filesystem::directory_iterator end;
	for (boost::filesystem::directory_iterator bus_iter(base_dir); bus_iter != end; ++bus_iter) {
		// バスでループ
		if (boost::filesystem::is_directory(*bus_iter)) {
			// ディレクトリのみ
			for (boost::filesystem::directory_iterator dev_iter(*bus_iter); dev_iter != end; ++dev_iter) {
				// バスに繋がってるデバイスでループ
				if (!boost::filesystem::is_directory(*dev_iter)) {
					// Friioであるかチェック
					if (is_friio(dev_iter->string())) {
						// ここまで来たらこのデバイスはfriioです。
						std::string this_device = bus_iter->leaf() + "/" + dev_iter->leaf();
						if (tuner_type_map.count(this_device) > 0) {
							if (target_type == tuner_type_map[this_device]) {
								friios.push_back(dev_iter->string());
							} else {
								std::cout << dev_iter->string() << " is "
								          << friioTypeName[tuner_type_map[this_device]] << ". skipped." << std::endl;
							}
						} else {
							std::cout << dev_iter->string() << " is Friio(Unknown). skipped." << std::endl;
						}
					}
				}
			}
		}
	}
	
	if (0 == friios.size()) {
		errx(1, "no friio found.");		
	}
	
	// 使用可能なFriioを探す
	int result_fd = -1;
	std::string result_devfile = "";
	
	std::cout << friios.size() << " friios found:" << std::endl;
	for (std::vector<std::string>::iterator iter = friios.begin(); iter != friios.end(); ++iter) {
		std::string devfile = *iter;
		
		std::cout << devfile << ": ";
		int fd = open(devfile.c_str(), O_RDWR);
		if (-1 == fd) {
			int errno_bak = errno;
			char buf[1024];
			memset(buf, '\0', sizeof(buf));
			std::cout << "open failed: " << strerror_r(errno_bak, buf, sizeof(buf)) << std::endl;
			continue;
		}
		
		if (-1 != result_fd) {
			// もう見つかっているのでclaimされているかのチェックだけ行う。
			
			usbdevfs_getdriver driver_info;
			memset(&driver_info, '\0', sizeof(driver_info));
			driver_info.interface = TARGET_INTERFACE;
			int r = ioctl(fd, USBDEVFS_GETDRIVER, &driver_info);
			if (r < 0) {
				int errno_bak = errno;
				
				if (errno_bak == ENODATA) { // 未使用の場合ここにくる。
					std::cout << "not used." << std::endl;
					close(fd);
					continue;
				}
				
				// エラー
				char buf[1024];
				memset(buf, '\0', sizeof(buf));
				std::cout << "getdriver failed: " << strerror_r(errno_bak, buf, sizeof(buf)) << std::endl;
				close(fd);
				continue;
			}
			
			std::cout << "used by " << driver_info.driver << std::endl;
			
			close(fd);
		} else {
			// まだ見つかってないのでclaimする。
			unsigned int intf = TARGET_INTERFACE;
			int r = ioctl(fd, USBDEVFS_CLAIMINTERFACE, &intf);
			if (r < 0) {
				int errno_bak = errno;				
				if (errno_bak == EBUSY) { // 使用中かチェック
					std::cout << "busy." << std::endl;
					close(fd);
					continue;
				}
			
				// エラー
				char buf[1024];
				memset(buf, '\0', sizeof(buf));
				std::cout << "claim failed: " << strerror_r(errno_bak, buf, sizeof(buf)) << std::endl;
				close(fd);
				continue;			
			}
	
			// 使用可能なfriioを発見
			result_fd = fd;
			result_devfile = devfile;
			std::cout << "use this friio." << std::endl;
		}
	}
	
	if (-1 == result_fd) {
		errx(1, "no friio can be used.");
	}
	
	std::cout << "device: " << result_devfile << std::endl;
	return result_fd;
}

bool is_friio(const std::string &devfile) {
	int f = open(devfile.c_str(), O_RDONLY);
	if (-1 == f) {
		if (errno == EACCES) {
			// 読み込み不可
			return false;
		}
		err(1, "can't open usbdevfile to read '%s'", devfile.c_str());
	}
	
	usb_device_descriptor usb_desc;
	memset(&usb_desc, '\0', sizeof(usb_desc));
	ssize_t rlen = read(f, &usb_desc, sizeof(usb_desc));
	if (-1 == rlen) {
		err(1, "can't read usbdevfile '%s'", devfile.c_str());
	}
	
	if (-1 == close(f)) {
		err(1, "can't close usbdevfile '%s'", devfile.c_str());
	}
	
	if (TARGET_ID_VENDOR == usb_desc.idVendor && TARGET_ID_PRODUCT == usb_desc.idProduct) {
		return true;
	} else {
		return false;
	}
}

// USB初期化処理
void UsbInitalize(int fd)
{
	try {
		// 固定処理Ａ(デバイス判定)
		FriioType type = UsbProcFixInitA(fd);
		if (FRIIO_BLACK != type) {
			throw "invalid tuner type."; // TODO:
		};
		
		// ?
		UsbProcBegin(fd);
		// ストリーム制御データ
		UsbProcStreamInit(fd);
		// ＬＥＤ制御(紫)
		UsbProcLED(fd, BON_LED_PURPLE);
		// 固定処理Ｂ
		UsbProcFixInitB(fd);
		// コントロールＬＥＤ制御(緑)
		UsbProcLED(fd, BON_LED_GREEN);
		// 初期化＼(＾o＾)／お疲れ様
	} catch(int err)
	{
		errx(1, "init failed.");
	}
}

// 処理開始(白黒で違うかも？)
// 便宜上Beginにしているが、本来どんなことをしているかは不明
static void UsbProcBegin(int fd)
{
	UsbSendVendorRequest(fd, 0x0000, 0x0000);
	UsbSendVendorRequest(fd, 0x000F, 0x0006);
	UsbSendVendorRequest(fd, 0x0000, 0x0000);
}

// 処理終了
// 便宜上Endにしている。恐らくバッファクリア
static void UsbProcEnd(int fd)
{
	UsbSendVendorRequest(fd, 0x0000, 0x0000);
}

// USB終了処理
static void UsbTerminate(int fd)
{
	UsbProcLED(fd, BON_LED_PURPLE);
	UsbProcEnd(fd);
}

// 固定処理Ａ
static FriioType UsbProcFixInitA(int fd)
{
	FriioType type;
	
	// STEP 1
	UsbSendVendorRequest(fd, 0x0002, 0x0011);
	// STEP 2
	UsbSendVendorRequest(fd, 0x0000, 0x0011);
	// STEP 3
	uint8_t buf1[] = "";
	buf1[0] = 0xC1;
	usbdevfs_ctrltransfer ctrl1;
	ctrl1.bRequestType = USB_DIR_OUT|USB_TYPE_VENDOR|USB_RECIP_DEVICE;
	ctrl1.bRequest     = 0x03;
	ctrl1.wValue       = 0x3000;
	ctrl1.wIndex       = 0x00FE;
	ctrl1.wLength      = 1;
	ctrl1.timeout      = REQUEST_TIMEOUT;
	ctrl1.data         = buf1;
	
	tuner_ctrl(fd, &ctrl1);

	// STEP 4
	uint8_t buf2[] = "";
	usbdevfs_ctrltransfer ctrl2;
	ctrl2.bRequestType = USB_DIR_IN|USB_TYPE_VENDOR|USB_RECIP_DEVICE;
	ctrl2.bRequest     = 0x02;
	ctrl2.wValue       = 0x3000;
	ctrl2.wIndex       = 0x0100;
	ctrl2.wLength      = 1;
	ctrl2.timeout      = REQUEST_TIMEOUT;
	ctrl2.data         = buf2;
	
	tuner_ctrl(fd, &ctrl2);
	
	if (buf2[0]) {
		// 白凡と判定
		type = FRIIO_WHITE;
	} else {
		// 黒凡と判定
		type = FRIIO_BLACK;
		// STEP4 Append
		uint8_t buf3[] = "";
		usbdevfs_ctrltransfer ctrl3;
		ctrl3.bRequestType = USB_DIR_IN|USB_TYPE_VENDOR|USB_RECIP_DEVICE;
		ctrl3.bRequest     = 0x02;
		ctrl3.wValue       = 0x3000;
		ctrl3.wIndex       = 0x0080;
		ctrl3.wLength      = 1;
		ctrl3.timeout      = REQUEST_TIMEOUT;
		ctrl3.data         = buf3;
		
		tuner_ctrl(fd, &ctrl3);
	}
	
	// STEP 5
	UsbSendVendorRequest(fd, 0x0004, 0x0030);
	
	return type;
}

// 固定処理Ｂ
static void UsbProcFixInitB(int fd)
{
	UsbSendVendorRequest(fd, 0x3040, 0x0001);
	UsbSendVendorRequest(fd, 0x3030, 0x0004);
	UsbSendVendorRequest(fd, 0x3040, 0x0005);
	UsbSendVendorRequest(fd, 0x3040, 0x0007);
	UsbSendVendorRequest(fd, 0x30ff, 0x000f);
	UsbSendVendorRequest(fd, 0x3027, 0x0011);
	UsbSendVendorRequest(fd, 0x302f, 0x0013);
	UsbSendVendorRequest(fd, 0x3031, 0x0014);
	UsbSendVendorRequest(fd, 0x3002, 0x0016);
	UsbSendVendorRequest(fd, 0x3020, 0x0022);
	UsbSendVendorRequest(fd, 0x3070, 0x0024);
	UsbSendVendorRequest(fd, 0x3079, 0x002c);
	UsbSendVendorRequest(fd, 0x3034, 0x002d);
	UsbSendVendorRequest(fd, 0x3028, 0x0030);
	UsbSendVendorRequest(fd, 0x3031, 0x0031);
	UsbSendVendorRequest(fd, 0x3001, 0x0038);
	UsbSendVendorRequest(fd, 0x3078, 0x0039);
	UsbSendVendorRequest(fd, 0x3090, 0x0048);
	UsbSendVendorRequest(fd, 0x3068, 0x0051);
	UsbSendVendorRequest(fd, 0x3038, 0x005e);
	UsbSendVendorRequest(fd, 0x3021, 0x00c0);
	UsbSendVendorRequest(fd, 0x3010, 0x00c1);
	UsbSendVendorRequest(fd, 0x301a, 0x00e4);
	UsbSendVendorRequest(fd, 0x301f, 0x00ea);
}

static void UsbProcLED(int fd, uint8_t bType)
{
	switch(bType)
	{
	case BON_LED_PURPLE:
		{
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x0002, 0x0000);
			UsbSendVendorRequest(fd, 0x0006, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x0002, 0x0000);
			UsbSendVendorRequest(fd, 0x0006, 0x0000);
			UsbSendVendorRequest(fd, 0x0002, 0x0000);
			UsbSendVendorRequest(fd, 0x0006, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x0002, 0x0000);
			UsbSendVendorRequest(fd, 0x0006, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x0002, 0x0000);
			UsbSendVendorRequest(fd, 0x0006, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x0002, 0x0000);
			UsbSendVendorRequest(fd, 0x0006, 0x0000);
			UsbSendVendorRequest(fd, 0x0002, 0x0000);
			UsbSendVendorRequest(fd, 0x0006, 0x0000);
			UsbSendVendorRequest(fd, 0x0002, 0x0000);
			UsbSendVendorRequest(fd, 0x0006, 0x0000);
			UsbSendVendorRequest(fd, 0x0002, 0x0000);
			UsbSendVendorRequest(fd, 0x0006, 0x0000);
			UsbSendVendorRequest(fd, 0x0002, 0x0000);
			UsbSendVendorRequest(fd, 0x0006, 0x0000);
			UsbSendVendorRequest(fd, 0x0002, 0x0000);
			UsbSendVendorRequest(fd, 0x0006, 0x0000);
			UsbSendVendorRequest(fd, 0x0002, 0x0000);
			UsbSendVendorRequest(fd, 0x0006, 0x0000);
			UsbSendVendorRequest(fd, 0x0002, 0x0000);
			UsbSendVendorRequest(fd, 0x0006, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x0008, 0x0000);
			UsbSendVendorRequest(fd, 0x000c, 0x0000);
		}
		break;
	case BON_LED_GREEN:
		{
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x0002, 0x0000);
			UsbSendVendorRequest(fd, 0x0006, 0x0000);
			UsbSendVendorRequest(fd, 0x0002, 0x0000);
			UsbSendVendorRequest(fd, 0x0006, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x0002, 0x0000);
			UsbSendVendorRequest(fd, 0x0006, 0x0000);
			UsbSendVendorRequest(fd, 0x0002, 0x0000);
			UsbSendVendorRequest(fd, 0x0006, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x0002, 0x0000);
			UsbSendVendorRequest(fd, 0x0006, 0x0000);
			UsbSendVendorRequest(fd, 0x0002, 0x0000);
			UsbSendVendorRequest(fd, 0x0006, 0x0000);
			UsbSendVendorRequest(fd, 0x0002, 0x0000);
			UsbSendVendorRequest(fd, 0x0006, 0x0000);
			UsbSendVendorRequest(fd, 0x0002, 0x0000);
			UsbSendVendorRequest(fd, 0x0006, 0x0000);
			UsbSendVendorRequest(fd, 0x0002, 0x0000);
			UsbSendVendorRequest(fd, 0x0006, 0x0000);
			UsbSendVendorRequest(fd, 0x0002, 0x0000);
			UsbSendVendorRequest(fd, 0x0006, 0x0000);
			UsbSendVendorRequest(fd, 0x0002, 0x0000);
			UsbSendVendorRequest(fd, 0x0006, 0x0000);
			UsbSendVendorRequest(fd, 0x0002, 0x0000);
			UsbSendVendorRequest(fd, 0x0006, 0x0000);
			UsbSendVendorRequest(fd, 0x0002, 0x0000);
			UsbSendVendorRequest(fd, 0x0006, 0x0000);
			UsbSendVendorRequest(fd, 0x0002, 0x0000);
			UsbSendVendorRequest(fd, 0x0006, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x0002, 0x0000);
			UsbSendVendorRequest(fd, 0x0006, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x0002, 0x0000);
			UsbSendVendorRequest(fd, 0x0006, 0x0000);
			UsbSendVendorRequest(fd, 0x0002, 0x0000);
			UsbSendVendorRequest(fd, 0x0006, 0x0000);
			UsbSendVendorRequest(fd, 0x000a, 0x0000);
			UsbSendVendorRequest(fd, 0x000e, 0x0000);
			UsbSendVendorRequest(fd, 0x0002, 0x0000);
			UsbSendVendorRequest(fd, 0x0006, 0x0000);
			UsbSendVendorRequest(fd, 0x0002, 0x0000);
			UsbSendVendorRequest(fd, 0x0006, 0x0000);
			UsbSendVendorRequest(fd, 0x0000, 0x0000);
			UsbSendVendorRequest(fd, 0x0004, 0x0000);
		}
		break;
	case BON_LED_RED:
		{
		}
		break;
	case BON_LED_WHITE:
		{
		}
		break;
	}
}

// ストリーム初期化処理
static void UsbProcStreamInit(int fd)
{
	UsbSendVendorRequest(fd, 0x0008, 0x0033);
	UsbSendVendorRequest(fd, 0x0040, 0x0037);
	UsbSendVendorRequest(fd, 0x001f, 0x003a);
	UsbSendVendorRequest(fd, 0x00ff, 0x003b);
	UsbSendVendorRequest(fd, 0x001f, 0x003c);
	UsbSendVendorRequest(fd, 0x00ff, 0x003d);
	UsbSendVendorRequest(fd, 0x0000, 0x0038);
	UsbSendVendorRequest(fd, 0x0000, 0x0035);
	UsbSendVendorRequest(fd, 0x0000, 0x0039);
	UsbSendVendorRequest(fd, 0x0000, 0x0036);
}
